from collections import defaultdict
from dataclasses import dataclass, field
from os.path import splitext

from xml_to_acts.node import Node


@dataclass
class Section:
    name: str
    params: dict = field(default_factory=dict)

    def tostring(self):
        params = "\n".join(f"{k}: {v}" for k, v in self.params.items())
        return f"[{self.name}]\n{params}"


def toACTS(root: Node, name: str):
    sections = []

    system = Section("System", {"Name": name})
    sections.append(system)

    params = defaultdict(list)
    for node in root.dfs():
        if node.is_class():
            key = f"{node.parent.value.strip()}(enum)"
            params[key].append(node.value.strip())
    params = {k: ",".join(v) for k, v in params.items()}
    parameter = Section("Parameter", params)
    sections.append(parameter)

    sections.append(Section("Constraint"))

    return "\n\n".join(s.tostring() for s in sections)


def write(root: Node, filename: str):
    name, _ = splitext(filename)
    with open(filename, "w+") as f:
        f.write(toACTS(root, name))



from dataclasses import dataclass, field
from typing import Optional

from pydrawio.mxgraphmodel import MxCell


@dataclass
class Node:
    id: str
    value: str
    style: dict[str, object]
    parent: Optional['Node'] = None
    children: list['Node'] = field(default_factory=list)

    def __repr__(self):
        return f"<{self.value}>" if self.is_class() else f"'{self.value}'"

    @classmethod
    def from_xml(cls, xml: MxCell):
        style = {}
        for entry in xml.style.strip().split(";"):
            key, *value = entry.strip().split("=", maxsplit=1)
            if value:
                style[key] = value[0]
            else:
                style[key] = True
        return cls(xml.id, xml.value, style)

    def is_root(self):
        return not self.parent and self.children

    def is_class(self):
        # return self.style.get("strokeColor", False) == "none"
        return not self.children

    def dfs(self):
        yield self
        for node in self.children:
            yield from node.dfs()

    def prettify(self, depth=0) -> str:
        res = "--" * depth + " " * bool(depth) + str(self)
        for node in self.children:
            res += "\n" + node.prettify(depth + 1)
        return res



from os.path import splitext, basename

import click

from xml_to_acts.drawio import xml_to_acts


@click.command()
@click.argument('input', type=click.Path(exists=True))
@click.argument('output', type=click.File("w"))
def main_cli(input, output):
    with open(input, "r", encoding="utf-8") as f:
        name, _ = splitext(input)
        result = xml_to_acts(f.read(), basename(name))
    output.write(result)



import logging
import xml.etree.ElementTree as ET

from pydrawio import decompress
from pydrawio.mxfile import Mxfile
from pydrawio.mxgraphmodel import MxGraphModel, MxCell

from xml_to_acts.acts import toACTS
from xml_to_acts.node import Node


class GraphException(Exception):
    pass


def hasattrs(obj, *attrs):
    return all(hasattr(obj, attr) for attr in attrs)


def is_edge(xml):
    return hasattrs(xml, "edge", "source", "target") and xml.edge


def is_node(xml):
    return hasattrs(xml, "value", "vertex") and xml.value


def create_graph(model: MxGraphModel) -> Node:
    nodes = {}
    edges = []
    for xml in model.content.items:
        if not isinstance(xml, MxCell):
            continue
        if is_node(xml):
            nodes[xml.id] = Node.from_xml(xml)
        if is_edge(xml):
            edges.append(xml)

    for xml in edges:
        parent, child = nodes[xml.source], nodes[xml.target]
        if child.parent and child.parent != parent:
            raise GraphException(f"Multiple parents for node {child} found: {parent} and {child.parent}")
        child.parent = parent
        parent.children.append(child)

    root = None
    for node in nodes.values():
        if node.parent or not node.children:
            continue
        if root and root != node:
            raise GraphException(f"Multiple root nodes found: {root} and {node}")
        else:
            root = node
    return root


def model_from_xml(xml):
    tree = ET.ElementTree(ET.fromstring(xml))
    root = tree.getroot()
    diagram = root[0]
    model_xml = ET.tostring(diagram[0], encoding="unicode")
    return MxGraphModel(model_xml)


def xml_to_acts(xml: str, name: str = "Convertor result"):
    mx_file = Mxfile(xml)

    diagram = mx_file.diagram[0]
    if diagram.value.strip():
        mx_graph_model = decompress(diagram.value)
    else:
        mx_graph_model = model_from_xml(xml)

    root = create_graph(mx_graph_model)
    logging.debug("\n" + root.prettify())
    logging.debug(f"Classes: {' '.join(map(str, filter(Node.is_class, root.dfs())))}")
    return toACTS(root, name)



from os.path import splitext, join

from xml_to_acts.drawio import xml_to_acts


def test_all_xml(xml_filename: str, xml_input_path: str, xml_output_path: str):
    xml_path = join(xml_input_path, xml_filename)
    with open(xml_path, "r", encoding="utf-8") as f:
        xml = f.read()

    acts = xml_to_acts(xml, "Convertor result")

    name, _ = splitext(xml_filename)
    xml_output_path = join(xml_output_path, f"{name}.txt")
    with open(xml_output_path, "r") as f:
        assert f.read().strip() == acts.strip()

